﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Dynamic;
using System.Dynamic;
using System.Reflection;
using Newtonsoft.Json;

namespace DynamicLinq
{
    class Program
    {
        static void Main(string[] args)
        {
            var list = InitData();
            List<string> DimensionList = new List<string>() { "Area", "Month" };
            string DynamicColumn = "Name";
            List<string> AllDynamicColumn = null;
            List<dynamic> result = DynamicLinq(list, DimensionList, DynamicColumn, out AllDynamicColumn);
            Console.WriteLine(JsonConvert.SerializeObject(result, Formatting.Indented));

            Console.WriteLine(JsonConvert.SerializeObject(list.Where("Name=@0", "张三"), Formatting.Indented)); 
            Console.Read();
        }

        private static List<House> InitData()
        {
            List<House> list = new List<House>();
            list.Add(new House
            {
                Name = "张三",
                Month = "2016-01",
                Area = "江夏区",
                DfMoney = 240.9,
                SfMoney = 30,
                RqfMoney = 25
            });

            list.Add(new House
            {
                Name = "张三",
                Month = "2016-02",
                Area = "江夏区",
                DfMoney = 167,
                SfMoney = 24.5,
                RqfMoney = 17.9
            });

            list.Add(new House
            {
                Name = "小燕子",
                Month = "2016-01",
                Area = "江夏区",
                DfMoney = 340.9,
                SfMoney = 20,
                RqfMoney = 55
            });

            list.Add(new House
            {
                Name = "小燕子",
                Month = "2016-02",
                Area = "江夏区",
                DfMoney = 67,
                SfMoney = 64.5,
                RqfMoney = 77.9
            });

            list.Add(new House
            {
                Name = "李四",
                Month = "2016-01",
                Area = "洪山区",
                DfMoney = 56.7,
                SfMoney = 24.7,
                RqfMoney = 13.2
            });

            list.Add(new House
            {
                Name = "李四",
                Month = "2016-02",
                Area = "洪山区",
                DfMoney = 65.2,
                SfMoney = 18.9,
                RqfMoney = 14.9
            });

            list.Add(new House
            {
                Name = "尔康",
                Month = "2016-01",
                Area = "洪山区",
                DfMoney = 156.7,
                SfMoney = 124.7,
                RqfMoney = 33.2
            });

            list.Add(new House
            {
                Name = "尔康",
                Month = "2016-02",
                Area = "洪山区",
                DfMoney = 35.2,
                SfMoney = 28.9,
                RqfMoney = 44.9
            });
            return list;
        }


        /// <summary>
        /// 动态Linq方式实现行转列
        /// </summary>
        /// <param name="list">数据</param>
        /// <param name="DimensionList">维度列</param>
        /// <param name="DynamicColumn">动态列</param>
        /// <returns>行转列后数据</returns>
        private static List<dynamic> DynamicLinq<T>(List<T> list, List<string> DimensionList, string DynamicColumn, out List<string> AllDynamicColumn) where T : class
        {
            //获取所有动态列
            var columnGroup = list.GroupBy(DynamicColumn, "new(it as Vm)") as IEnumerable<IGrouping<dynamic, dynamic>>;
            List<string> AllColumnList = new List<string>();
            foreach (var item in columnGroup)
            {
                if (!string.IsNullOrEmpty(item.Key))
                {
                    AllColumnList.Add(item.Key);
                }
            }
            AllDynamicColumn = AllColumnList;
            var dictFunc = new Dictionary<string, Func<T, bool>>();
            foreach (var column in AllColumnList)
            {
                var func = DynamicExpression.ParseLambda<T, bool>(string.Format("{0}==\"{1}\"", DynamicColumn, column)).Compile();
                dictFunc[column] = func;
            }

            //获取实体所有属性
            Dictionary<string, PropertyInfo> PropertyInfoDict = new Dictionary<string, PropertyInfo>();
            Type type = typeof(T);
            var propertyInfos = type.GetProperties(BindingFlags.Instance | BindingFlags.Public);
            //数值列
            List<string> AllNumberField = new List<string>();
            foreach (var item in propertyInfos)
            {
                PropertyInfoDict[item.Name] = item;
                if (item.PropertyType == typeof(int) || item.PropertyType == typeof(double) || item.PropertyType == typeof(float))
                {
                    AllNumberField.Add(item.Name);
                }
            }

            //分组
            var dataGroup = list.GroupBy(string.Format("new ({0})", string.Join(",", DimensionList)), "new(it as Vm)") as IEnumerable<IGrouping<dynamic, dynamic>>;
            List<dynamic> listResult = new List<dynamic>();
            IDictionary<string, object> itemObj = null;
            T vm2 = default(T);
            foreach (var group in dataGroup)
            {
                itemObj = new ExpandoObject();
                var listVm = group.Select(e => e.Vm as T).ToList();
                //维度列赋值
                vm2 = listVm.FirstOrDefault();
                foreach (var key in DimensionList)
                {
                    itemObj[key] = PropertyInfoDict[key].GetValue(vm2);
                }

                foreach (var column in AllColumnList)
                {
                    vm2 = listVm.FirstOrDefault(dictFunc[column]);
                    if (vm2 != null)
                    {
                        foreach (string name in AllNumberField)
                        {
                            itemObj[name + column] = PropertyInfoDict[name].GetValue(vm2);
                        }
                    }
                }
                listResult.Add(itemObj);
            }
            return listResult;
        }
    }
}
