﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Dynamic;
using System.Dynamic;
using System.Reflection;
using Newtonsoft.Json;
using System.Data;
using System.IO;
using System.Diagnostics;

namespace Test
{
    class Program
    {
        static void Main(string[] args)
        {
            Console.WriteLine("准备数据...");
            DataTable table = InitDataTable();
            List<Model> entitys = InitEntitys(table);
            Stopwatch stopWatch = new Stopwatch();

            Console.WriteLine("开始测试...");
            stopWatch.Start();
            for (int i = 0; i < 1; i++)
            {
                //List行转列 测试
                List<string> DimensionList = new List<string>() { "住院号", "姓名" };
                string DynamicColumn = "费用类别";
                List<string> AllDynamicColumn = null;
                List<dynamic> result = DynamicLinq(entitys, DimensionList, DynamicColumn, out AllDynamicColumn);
            }
            stopWatch.Stop();
            Console.WriteLine($"List行转列时间：{stopWatch.Elapsed.TotalMilliseconds}");

            stopWatch.Restart();
            for (int i = 0; i < 1; i++)
            {
                List<string> DimensionList = new List<string>() { "住院号", "姓名" };
                string DynamicColumn = "费用类别";
                List<string> AllDynamicColumn = null;
                DataTable dtResult = RowToColumn(table, DimensionList, DynamicColumn, out AllDynamicColumn);
            }
            Console.WriteLine($"DataTable行转列时间：{stopWatch.Elapsed.TotalMilliseconds}");
            Console.ReadKey();
        }

        /// <summary>
        /// 动态Linq方式实现行转列
        /// </summary>
        /// <param name="list">数据</param>
        /// <param name="DimensionList">维度列</param>
        /// <param name="DynamicColumn">动态列</param>
        /// <returns>行转列后数据</returns>
        private static List<dynamic> DynamicLinq<T>(List<T> list, List<string> DimensionList, string DynamicColumn, out List<string> AllDynamicColumn) where T : class
        {
            //获取所有动态列
            var columnGroup = list.GroupBy(DynamicColumn, "new(it as Vm)") as IEnumerable<IGrouping<dynamic, dynamic>>;
            List<string> AllColumnList = new List<string>();
            foreach (var item in columnGroup)
            {
                if (!string.IsNullOrEmpty(item.Key))
                {
                    AllColumnList.Add(item.Key);
                }
            }
            AllDynamicColumn = AllColumnList;
            var dictFunc = new Dictionary<string, Func<T, bool>>();
            foreach (var column in AllColumnList)
            {
                var func = DynamicExpression.ParseLambda<T, bool>(string.Format("{0}==\"{1}\"", DynamicColumn, column)).Compile();
                dictFunc[column] = func;
            }

            //获取实体所有属性
            Dictionary<string, PropertyInfo> PropertyInfoDict = new Dictionary<string, PropertyInfo>();
            Type type = typeof(T);
            var propertyInfos = type.GetProperties(BindingFlags.Instance | BindingFlags.Public);
            //数值列
            List<string> AllNumberField = new List<string>();
            foreach (var item in propertyInfos)
            {
                PropertyInfoDict[item.Name] = item;
                if (item.PropertyType == typeof(int) || item.PropertyType == typeof(double) || item.PropertyType == typeof(float))
                {
                    AllNumberField.Add(item.Name);
                }
            }

            //分组
            var dataGroup = list.GroupBy(string.Format("new ({0})", string.Join(",", DimensionList)), "new(it as Vm)") as IEnumerable<IGrouping<dynamic, dynamic>>;
            List<dynamic> listResult = new List<dynamic>();
            IDictionary<string, object> itemObj = null;
            T vm2 = default(T);
            foreach (var group in dataGroup)
            {
                itemObj = new ExpandoObject();
                var listVm = group.Select(e => e.Vm as T).ToList();
                //维度列赋值
                vm2 = listVm.FirstOrDefault();
                foreach (var key in DimensionList)
                {
                    itemObj[key] = PropertyInfoDict[key].GetValue(vm2);
                }

                foreach (var column in AllColumnList)
                {
                    vm2 = listVm.FirstOrDefault(dictFunc[column]);
                    if (vm2 != null)
                    {
                        foreach (string name in AllNumberField)
                        {
                            itemObj[name + column] = PropertyInfoDict[name].GetValue(vm2);
                        }
                    }
                }
                listResult.Add(itemObj);
            }
            return listResult;
        }


        /// <summary>
        /// 动态Linq方式实现行转列
        /// </summary>
        /// <param name="list">数据</param>
        /// <param name="DimensionList">维度列</param>
        /// <param name="DynamicColumn">动态列</param>
        /// <returns>行转列后数据</returns>
        private static DataTable RowToColumn(DataTable dt, List<string> DimensionList, string DynamicColumn, out List<string> AllDynamicColumn)
        {
            //获取所有动态列
            AllDynamicColumn = new List<string>();
            foreach (DataRow dr in dt.DefaultView.ToTable(true, DynamicColumn).Rows)
            {
                if (dr[DynamicColumn] != null && !string.IsNullOrEmpty(dr[DynamicColumn].ToString()))
                {
                    AllDynamicColumn.Add(dr[DynamicColumn].ToString());
                }
            }

            //数值列
            Dictionary<string, Type> AllNumberColumn = new Dictionary<string, Type>();
            foreach (DataColumn item in dt.Columns)
            {
                if (item.DataType == typeof(int) || item.DataType == typeof(double) || item.DataType == typeof(float))
                {
                    AllNumberColumn.Add(item.ColumnName, item.DataType);
                }
            }

            //结果DataTable创建
            DataTable dtResult = new DataTable();
            foreach (var item in DimensionList)
            {
                dtResult.Columns.Add(item, typeof(string));
            }
            //动态列
            foreach (var dynamicValue in AllDynamicColumn)
            {
                foreach (var item in AllNumberColumn.Keys)
                {
                    dtResult.Columns.Add(item + dynamicValue, AllNumberColumn[item]);
                }
            }

            //分组-优化性能
            Dictionary<string, List<DataRow>> dict = new Dictionary<string, List<DataRow>>();
            List<DataRow> drList = null;
            string groupKey = "";
            foreach (DataRow dr in dt.Rows)
            {
                groupKey = "";
                foreach (var item in DimensionList)
                {
                    groupKey += dr[item] + "#";
                }
                if (!dict.TryGetValue(groupKey, out drList))
                {
                    drList = new List<DataRow>();
                    dict[groupKey] = drList;
                }
                drList.Add(dr);
            }

            DataRow drReult = null;
            DataTable dtTemp = null;
            Dictionary<object, DataTable> dictTable = null;
            foreach (var kv in dict)
            {
                drReult = dtResult.NewRow();
                var arrKey = kv.Key.Split('#');
                int i = 0;
                foreach (var key in DimensionList)
                {
                    drReult[key] = arrKey[i];
                    i++;
                }
                dictTable = (from p in kv.Value.AsEnumerable()
                             group p by p.Field<object>(DynamicColumn) into g
                             select g).ToDictionary(e => e.Key, e => e.CopyToDataTable());
                foreach (var dynamicValue in AllDynamicColumn)
                {
                    if (dictTable.TryGetValue(dynamicValue, out dtTemp))
                    {
                        foreach (var numColumn in AllNumberColumn.Keys)
                        {
                            drReult[numColumn + dynamicValue] = dtTemp.Compute("sum(" + numColumn + ")", "");
                        }
                    }
                    else
                    {
                        foreach (var numColumn in AllNumberColumn.Keys)
                        {
                            drReult[numColumn + dynamicValue] = 0;
                        }
                    }
                }
                dtResult.Rows.Add(drReult);
            }
            return dtResult;
        }


        private static DataTable InitDataTable()
        {
            var ds = new DataSet();
            ds.ReadXml(Path.Combine(System.Windows.Forms.Application.StartupPath, "data.xml"), XmlReadMode.Auto);
            DataTable dt = ds.Tables[0].Copy();
            for (int i = 0; i < 5; i++)
            {
                foreach (DataRow row in dt.Rows)
                {
                    var item = new object[8];
                    item[0] = row["住院号"].ToString() + i;
                    item[1] = row["姓名"].ToString() + i;
                    item[2] = null;
                    item[3] = null;
                    item[4] = row["住院科室"].ToString() + i;
                    item[5] = row["费用类别"].ToString() + i;
                    item[6] = Convert.ToInt32(row["开单数"]) + i;
                    item[7] = Convert.ToDecimal(row["金额"]) + i;
                    ds.Tables[0].LoadDataRow(item, true);
                }
            }
            return ds.Tables[0];
        }

        private static List<Model> InitEntitys(DataTable dt)
        {
            List<Model> entitys = new List<Model>();

            foreach (DataRow row in dt.Rows)
            {
                Model entity = new Model();
                entity.住院号 = row["住院号"].ToString();
                entity.住院科室 = row["住院科室"].ToString();
                entity.姓名 = row["姓名"].ToString();
                entity.开单数 = Convert.ToInt32(row["开单数"]);
                entity.费用类别 = row["费用类别"].ToString();
                entity.金额 = Convert.ToDecimal(row["金额"]);
                entitys.Add(entity);
            }
            return entitys;
        }

    }
}
